/* Native-dependent code for GNU/Linux SPARC.

   Copyright 2001, 2002 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "regcache.h"
#include "gdbcore.h"
#include "gdb_string.h"
#include "inferior.h"
#include "target.h"
   
#include <sys/procfs.h>
#include <sys/ptrace.h>

#include <asm/unistd.h>

/* Prototypes for supply_gregset etc.  */
#include "gregset.h"

typedef unsigned long long elf64_greg_t;
typedef elf64_greg_t elf64_gregset_t[36];
typedef unsigned int elf32_greg_t;
typedef elf32_greg_t elf32_gregset_t[38];

static elf64_greg_t
tstate_to_ccr(elf64_greg_t tstate)
{
	return (tstate >> 32) & 0xff;
}

static elf64_greg_t
ccr_to_tstate(elf64_greg_t ccr)
{
	return (ccr & 0xffULL) << 32;
}

static void
supply_unobtainable_sparc64_regs (void)
{
  /* ??? We could be clever and derive these from other sources
     ??? such as tstate, but that causes testsuite failures
     ??? because things like sparc_pop_frame would need to do such
     ??? derivations too.  */
  supply_register (FPRS_REGNUM, NULL);
  supply_register (ASI_REGNUM, NULL);
  supply_register (VER_REGNUM, NULL);
  supply_register (TICK_REGNUM, NULL);
  supply_register (PIL_REGNUM, NULL);
  supply_register (PSTATE_REGNUM, NULL);
  supply_register (TSTATE_REGNUM, NULL);
  supply_register (TBA_REGNUM, NULL);
  supply_register (TL_REGNUM, NULL);
  supply_register (TT_REGNUM, NULL);
  supply_register (TPC_REGNUM, NULL);
  supply_register (TNPC_REGNUM, NULL);
  supply_register (WSTATE_REGNUM, NULL);
  supply_register (CWP_REGNUM, NULL);
  supply_register (CANSAVE_REGNUM, NULL);
  supply_register (CANRESTORE_REGNUM, NULL);
  supply_register (CLEANWIN_REGNUM, NULL);
  supply_register (OTHERWIN_REGNUM, NULL);
  supply_register (ASR16_REGNUM, NULL);
  supply_register (ASR17_REGNUM, NULL);
  supply_register (ASR18_REGNUM, NULL);
  supply_register (ASR19_REGNUM, NULL);
  supply_register (ASR20_REGNUM, NULL);
  supply_register (ASR21_REGNUM, NULL);
  supply_register (ASR22_REGNUM, NULL);
  supply_register (ASR23_REGNUM, NULL);
  supply_register (ASR24_REGNUM, NULL);
  supply_register (ASR25_REGNUM, NULL);
  supply_register (ASR26_REGNUM, NULL);
  supply_register (ASR27_REGNUM, NULL);
  supply_register (ASR28_REGNUM, NULL);
  supply_register (ASR29_REGNUM, NULL);
  supply_register (ASR30_REGNUM, NULL);
  supply_register (ASR31_REGNUM, NULL);
  supply_register (ICC_REGNUM, NULL);
  supply_register (XCC_REGNUM, NULL);
  supply_register (FCC0_REGNUM, NULL);
  supply_register (FCC1_REGNUM, NULL);
  supply_register (FCC2_REGNUM, NULL);
  supply_register (FCC3_REGNUM, NULL);
}

void
supply_gregset (gdb_gregset_t *__gregsetp)
{
  if (GDB_TARGET_IS_SPARC64)
    {
      elf64_greg_t *regp = (elf64_greg_t *) __gregsetp;
      elf64_greg_t ccr;
      int i;

      for (i = G0_REGNUM; i <= I7_REGNUM; i++)
	supply_register (i, (char *) (regp + (i - G0_REGNUM)));

      ccr = tstate_to_ccr(regp[32]);
      supply_register (CCR_REGNUM, (char *) &ccr);
      supply_register (PC_REGNUM, (char *) (regp + 33));
      supply_register (NPC_REGNUM, (char *) (regp + 34));
      supply_register (Y_REGNUM, (char *) (regp + 35));

      supply_unobtainable_sparc64_regs ();
    }
  else
    {
      elf32_greg_t *regp = (elf32_greg_t *) __gregsetp;
      int i;

      for (i = G0_REGNUM; i <= I7_REGNUM; i++)
	supply_register (i, (char *) (regp + (i - G0_REGNUM)));

      supply_register (PS_REGNUM, (char *) (regp + 32));

      supply_register (PC_REGNUM, (char *) (regp + 33));
      supply_register (NPC_REGNUM, (char *) (regp + 34));
      supply_register (Y_REGNUM, (char *) (regp + 35));

      /* Fill inaccessible registers with zero.  */
      supply_register (WIM_REGNUM, NULL);
      supply_register (TBR_REGNUM, NULL);
      supply_register (CPS_REGNUM, NULL);
    }
}

void
fill_gregset (gdb_gregset_t *__gregsetp, int regno)
{
  int i;

  if (GDB_TARGET_IS_SPARC64)
    {
      elf64_greg_t *regp = (elf64_greg_t *) __gregsetp;
      elf64_greg_t ccr;

      for (i = G0_REGNUM; i <= I7_REGNUM; i++)
	if (regno == -1 || regno == i)
	  regcache_collect (i, regp + (i - G0_REGNUM));

      if (regno == -1 || regno == CCR_REGNUM)
	{
	  regcache_collect (CCR_REGNUM, &ccr);
	  regp[32] = ccr_to_tstate(ccr);
	}

      if (regno == -1 || regno == PC_REGNUM)
	regcache_collect (PC_REGNUM, regp + 33);
      if (regno == -1 || regno == NPC_REGNUM)
	regcache_collect (NPC_REGNUM, regp + 34);
      if (regno == -1 || regno == Y_REGNUM)
	regcache_collect (Y_REGNUM, regp + 35);
    }
  else
    {
      elf32_greg_t *regp = (elf32_greg_t *) __gregsetp;

      for (i = G0_REGNUM; i <= I7_REGNUM; i++)
	if (regno == -1 || regno == i)
	  regcache_collect (i, regp + (i - G0_REGNUM));

      if (regno == -1 || regno == PS_REGNUM)
	regcache_collect (PS_REGNUM, regp + 32);

      if (regno == -1 || regno == PC_REGNUM)
	regcache_collect (PC_REGNUM, regp + 33);
      if (regno == -1 || regno == NPC_REGNUM)
	regcache_collect (NPC_REGNUM, regp + 34);
      if (regno == -1 || regno == Y_REGNUM)
	regcache_collect (Y_REGNUM, regp + 35);
    }
}

typedef struct {
  unsigned long long pr_regs[32];
  unsigned long long pr_fsr;
  unsigned long long pr_gsr;
  unsigned long long pr_fprs;
} elf64_fpregset_t;

typedef struct {
  unsigned int pr_regs[32];
  unsigned int __unused;
  unsigned int pr_fsr;
  unsigned int pr_qcnt;
  unsigned int pr_q_entrysize;
  unsigned int pr_en;
  unsigned int pr_q[64];
} elf32_fpregset_t;

void
supply_fpregset (gdb_fpregset_t *__fpregsetp)
{
  int i;

  if (GDB_TARGET_IS_SPARC64)
    {
      elf64_fpregset_t *regs = (elf64_fpregset_t *) __fpregsetp;
      unsigned int *fp_single = (unsigned int *) &regs->pr_regs[0];
      unsigned long long *fp_double = (unsigned long long *) &fp_single[32];

      for (i = 0; i < 32; i++)
	supply_register (i + FP0_REGNUM, (char *) &fp_single[i]);
      for (i = 0; i < 16; i++)
	supply_register (i + FP0_REGNUM + 32, (char *) &fp_double[i]);

      /* XXX No GSR_REGNUM */
      supply_register (FPRS_REGNUM, (char *) &regs->pr_fprs);
      supply_register (FSR_REGNUM, (char *) &regs->pr_fsr);
    }
  else
    {
      elf32_fpregset_t *regs = (elf32_fpregset_t *) __fpregsetp;

      for (i = FP0_REGNUM; i < FP0_REGNUM + 32; i++)
	supply_register (i, (char *) &regs->pr_regs[i - FP0_REGNUM]);
      supply_register (FPS_REGNUM, (char *) &regs->pr_fsr);
    }
}

void
fill_fpregset (gdb_fpregset_t *__fpregsetp, int regno)
{
  int i;

  if (GDB_TARGET_IS_SPARC64)
    {
      elf64_fpregset_t *regs = (elf64_fpregset_t *) __fpregsetp;
      unsigned int *fp_single = (unsigned int *) &regs->pr_regs[0];
      unsigned long long *fp_double = (unsigned long long *) &fp_single[32];

      for (i = FP0_REGNUM; i < FP0_REGNUM + 32; i++)
	if (regno == -1 || regno == i)
	  regcache_collect (i, (char *) &fp_single[i - FP0_REGNUM]);
      for (i = FP0_REGNUM + 32; i < FP0_REGNUM + 32 + 16; i++)
	if (regno == -1 || regno == i)
	  regcache_collect (i, (char *) &fp_double[i - (FP0_REGNUM + 32)]);

      /* XXX No GSR_REGNUM */

      if (regno == -1 || regno == FPRS_REGNUM)
	regcache_collect (FPRS_REGNUM, &regs->pr_fprs);
      if (regno == -1 || regno == FSR_REGNUM)
	regcache_collect (FSR_REGNUM, &regs->pr_fsr);
    }
  else
    {
      elf32_fpregset_t *regs = (elf32_fpregset_t *) __fpregsetp;

      for (i = FP0_REGNUM; i < FP0_REGNUM + 32; i++)
	if (regno == -1 || regno == i)
	  regcache_collect (i, &regs->pr_regs[i - FP0_REGNUM]);

      if (regno == -1 || regno == FPS_REGNUM)
	regcache_collect (FPS_REGNUM, &regs->pr_fsr);
    }
}

/*  Use a local version of this function to get the correct types for
    regsets, until multi-arch core support is ready.  */

static void
fetch_core_registers (char *core_reg_sect, unsigned core_reg_size,
		      int which, CORE_ADDR reg_addr)
{

  if (which == 0)
    {
      if (GDB_TARGET_IS_SPARC64)
	{
	  elf64_gregset_t gregset;

	  memcpy ((char *) &gregset, core_reg_sect, sizeof (gregset));
	  supply_gregset ((gdb_gregset_t *) &gregset);
	}
      else
	{
	  elf32_gregset_t gregset;

	  memcpy ((char *) &gregset, core_reg_sect, sizeof (gregset));
	  supply_gregset ((gdb_gregset_t *) &gregset);
	}
    }
  else if (which == 2)
    {
      if (GDB_TARGET_IS_SPARC64)
	{
	  elf64_fpregset_t fpregset;

	  memcpy ((char *) &fpregset, core_reg_sect, sizeof (fpregset));
	  supply_fpregset ((gdb_fpregset_t *) &fpregset);
	}
      else
	{
	  elf32_fpregset_t fpregset;

	  memcpy ((char *) &fpregset, core_reg_sect, sizeof (fpregset));
	  supply_fpregset ((gdb_fpregset_t *) &fpregset);
	}
    }
}

/* Size of a general (integer) register: */
extern int sparc_intreg_size (void);
#define SPARC_INTREG_SIZE (sparc_intreg_size ())

static elf32_greg_t
tstate_to_psr(elf64_greg_t tstate)
{
  return ((tstate & 0x1f) |
	  0x80 |
	  ((tstate & (elf64_greg_t)0x0f00000000) >> 12) |
	  ((tstate & (elf64_greg_t)0xf000000000) >> 20) |
	  0xff000000);
}

static elf64_greg_t
psr_to_tstate_icc(elf32_greg_t psr)
{
  elf64_greg_t tstate = ((elf64_greg_t)(psr & 0x00f00000)) << 12;

  if ((psr & (0xffffffff)) == 0xff000000)
    tstate |= ((elf64_greg_t)(psr & 0x000f0000)) << 20;

  return tstate;
}

struct pt64_regs
{
  elf64_greg_t	u_regs[16];
  elf64_greg_t	tstate;
  elf64_greg_t	tpc;
  elf64_greg_t	tnpc;
  elf32_greg_t	y;
  elf32_greg_t	fprs;
};

struct pt32_regs
{
  elf32_greg_t	psr;
  elf32_greg_t	pc;
  elf32_greg_t	npc;
  elf32_greg_t	y;
  elf32_greg_t	u_regs[16];
};

#define UREG_G1        0
#define UREG_G2        1
#define UREG_G3        2
#define UREG_G4        3
#define UREG_G5        4
#define UREG_G6        5
#define UREG_G7        6
#define UREG_O0        7
#define UREG_O1        8
#define UREG_O2        9
#define UREG_O3        10
#define UREG_O4        11
#define UREG_O5        12
#define UREG_O6        13
#define UREG_O7        14

struct pt64_fpregs
{
  elf32_greg_t	regs[64];
  elf64_greg_t	fsr;
};

struct pt32_fpregs
{
  elf32_greg_t	regs[32];
  elf32_greg_t	fsr;
  elf32_greg_t	flags;
  elf32_greg_t	extra;
  elf32_greg_t	fpqd;
  struct fpu_fq {
    elf32_greg_t	insnaddr;
    elf32_greg_t	insn;
  } fpq[16];
};

void
fetch_inferior_registers (int regno)
{
  int intreg_size = SPARC_INTREG_SIZE;
  int proc_id, i;
  char *buf;

  /* GNU/Linux LWP ID's are process ID's.  */
  if ((proc_id = TIDGET (inferior_ptid)) == 0)
    proc_id = PIDGET (inferior_ptid);		/* Not a threaded program.  */

  if (sizeof (void *) == 8)
    {
      union {
	struct pt64_regs regs;
	struct pt64_fpregs fpregs;
      } u;
      int offset = (intreg_size == 4) ? 4 : 0;

      if (regno <= O7_REGNUM
	  || (regno >= PC_REGNUM && regno != FSR_REGNUM)
	  || (!register_cached (SP_REGNUM) && regno <= I7_REGNUM))
	{
	  if (0 != ptrace (PTRACE_GETREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.regs, 0))
	    perror ("ptrace_getregs");

	  supply_register (G0_REGNUM, NULL);
	  if (regno == -1)
	    {
	      for (i = G1_REGNUM; i < G1_REGNUM + 15; i++)
		{
		  char *regptr = (char *)&u.regs.u_regs[i - G1_REGNUM + UREG_G1];

		  regptr += offset;
		  supply_register (i, regptr);
		}
	    }
	  else if (regno >= G1_REGNUM && regno < G1_REGNUM + 15)
	    {
	      char *regptr = (char *)&u.regs.u_regs[regno - G1_REGNUM + UREG_G1];

	      regptr += offset;
	      supply_register (regno, regptr);
	    }
	  else if (regno >= L0_REGNUM && regno <= I7_REGNUM)
	    {
	      char *regptr = (char *)&u.regs.u_regs[SP_REGNUM - G1_REGNUM + UREG_G1];

	      regptr += offset;
	      supply_register (SP_REGNUM, regptr);
	    }

	  if (regno == -1 ||
	      (! GDB_TARGET_IS_SPARC64 && regno == PS_REGNUM) ||
	      (GDB_TARGET_IS_SPARC64 && regno == CCR_REGNUM))
	    {
	      if (! GDB_TARGET_IS_SPARC64)
		{
		  elf32_greg_t psr_convert = tstate_to_psr(u.regs.tstate);
		  supply_register (PS_REGNUM, (char *) &psr_convert);
		}
	      else
		{
		  elf64_greg_t ccr_convert = tstate_to_ccr(u.regs.tstate);
		  supply_register (CCR_REGNUM, (char *) &ccr_convert);
		}
	    }
	  if (regno == -1 || regno == PC_REGNUM)
	    supply_register (PC_REGNUM, ((char *) &u.regs.tpc) + offset);
	  if (regno == -1 || regno == NPC_REGNUM)
	    supply_register (NPC_REGNUM, ((char *) &u.regs.tnpc) + offset);
	  if (regno == -1 || regno == Y_REGNUM)
	    supply_register (Y_REGNUM, ((char *) &u.regs.y) + offset);

	  if (regno == -1)
	    {
	      if (! GDB_TARGET_IS_SPARC64)
		{
		  supply_register (WIM_REGNUM, NULL);
		  supply_register (TBR_REGNUM, NULL);
		  supply_register (CPS_REGNUM, NULL);
		}
	      else
		{
		  supply_unobtainable_sparc64_regs ();
		}
	    }
	}

      if (regno == -1 ||
	  (! GDB_TARGET_IS_SPARC64 && regno == FPS_REGNUM) ||
	  (GDB_TARGET_IS_SPARC64   && regno == FSR_REGNUM) ||
	  (regno >= FP0_REGNUM
	   && regno < FP0_REGNUM + ((GDB_TARGET_IS_SPARC64) ? 48 : 32)))
	{
	  elf64_greg_t *fp_double = (elf64_greg_t *)&u.fpregs.regs[32];

	  if (0 != ptrace (PTRACE_GETFPREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	    perror ("ptrace_getfpregs");

	  if (regno == -1)
	    {
	      if (GDB_TARGET_IS_SPARC64)
		{
		  for (i = 0; i < 32; i++)
		    supply_register (FP0_REGNUM + i, (char *) &u.fpregs.regs[i]);
		  for (i = 0; i < 16; i++)
		    supply_register (FP0_REGNUM + i + 32, (char *) &fp_double[i]);
		  supply_register (FSR_REGNUM, (char *) &u.fpregs.fsr);
		}
	      else
		{
		  for (i = 0; i < 32; i++)
		    supply_register (FP0_REGNUM + i, (char *)&u.fpregs.regs[i]);
		  supply_register (FPS_REGNUM, ((char *) &u.fpregs.fsr) + offset);
		}
	    }
	  else
	    {
	      if (regno >= FP0_REGNUM && regno < (FP0_REGNUM + 32))
		supply_register (regno, (char *) &u.fpregs.regs[regno - FP0_REGNUM]);
	      else if (regno >= FP0_REGNUM + 32
		       && regno < FP0_REGNUM + 48)
		supply_register (regno, (char *) &fp_double[regno - 32 - FP0_REGNUM]);
	      else if (GDB_TARGET_IS_SPARC64)
		supply_register (FSR_REGNUM, (char *) &u.fpregs.fsr);
	      else
		supply_register (FPS_REGNUM, ((char *) &u.fpregs.fsr) + offset);
	    }
	}
    }
  else
    {
      union {
	struct pt32_regs regs;
	struct pt32_fpregs fpregs;
      } u;

      if (regno <= O7_REGNUM
	  || regno >= Y_REGNUM
	  || (!register_cached (SP_REGNUM) && regno <= I7_REGNUM))
	{
	  if (0 != ptrace (PTRACE_GETREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.regs, 0))
	    perror ("ptrace_getregs");

	  supply_register (G0_REGNUM, NULL);
	  if (regno == -1)
	    {
	      for (i = G1_REGNUM; i < G1_REGNUM + 15; i++)
		{
		  char *regptr = (char *)&u.regs,u_regs[i - G1_REGNUM + UREG_G1];

		  supply_register (i, regptr);
		}
	    }
	  else if (regno >= G1_REGNUM && regno < G1_REGNUM + 15)
	    {
	      char *regptr = (char *)&u.regs.u_regs[regno - G1_REGNUM + UREG_G1];

	      supply_register (regno, regptr);
	    }
	  else if (regno >= L0_REGNUM && regno <= I7_REGNUM)
	    {
	      char *regptr = (char *)&u.regs.u_regs[SP_REGNUM - G1_REGNUM + UREG_G1];

	      supply_register (SP_REGNUM, regptr);
	    }

	  if (regno == -1 || regno == PS_REGNUM)
	    supply_register (PS_REGNUM, (char *) &u.regs.psr);
	  if (regno == -1 || regno == PC_REGNUM)
	    supply_register (PC_REGNUM, (char *) &u.regs.pc);
	  if (regno == -1 || regno == NPC_REGNUM)
	    supply_register (NPC_REGNUM, (char *) &u.regs.npc);
	  if (regno == -1 || regno == Y_REGNUM)
	    supply_register (Y_REGNUM, (char *) &u.regs.y);

	  if (regno == -1)
	    {
	      supply_register (WIM_REGNUM, NULL);
	      supply_register (TBR_REGNUM, NULL);
	      supply_register (CPS_REGNUM, NULL);
	    }
	}

      if (regno == -1 ||
	  regno == FPS_REGNUM ||
	  (regno >= FP0_REGNUM
	   && regno < FP0_REGNUM + 32))
	{
	  if (0 != ptrace (PTRACE_GETFPREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	    perror ("ptrace_getfpregs");

	  if (regno == -1)
	    {
	      for (i = 0; i < 32; i++)
		supply_register (FP0_REGNUM + i, (char *)&u.fpregs.regs[i]);
	      supply_register (FPS_REGNUM, (char *) &u.fpregs.fsr);
	    }
	  else if (regno >= FP0_REGNUM && regno < FP0_REGNUM + 32)
	    {
	      supply_register (regno, (char *)&u.fpregs.regs[regno - FP0_REGNUM]);
	    }
	  else
	    supply_register (FPS_REGNUM, (char *) &u.fpregs.fsr);
	}
    }

  if (regno == -1)
    {
      buf = alloca (16 * intreg_size);
      target_read_memory (read_sp (), buf, 16 * intreg_size);
      for (i = 0; i <= (I7_REGNUM - L0_REGNUM); i++)
	supply_register (i + L0_REGNUM,
			 buf + (i * intreg_size));
    }
  else if (regno >= L0_REGNUM && regno <= I7_REGNUM)
    {
      buf = alloca (intreg_size);
      target_read_memory (read_sp () + REGISTER_BYTE (regno) - REGISTER_BYTE (L0_REGNUM),
			  buf, intreg_size);
      supply_register (regno, buf);
    }
}

static void
store_all_registers (int proc_id)
{
  int i;

  if (sizeof (void *) == 8)
    {
      union {
	struct pt64_regs regs;
	struct pt64_fpregs fpregs;
      } u;
      int offset = (SPARC_INTREG_SIZE == 4) ? 4 : 0;

      for (i = G1_REGNUM; i < G0_REGNUM + 16; i++)
	{
	  char *regptr = ((char *)&u.regs.u_regs[i - G1_REGNUM + UREG_G1]);

	  if (offset != 0)
	    memset (regptr, 0, sizeof(elf64_greg_t));
	  regptr += offset;
	  read_register_gen (i, regptr);
	}
      if (! GDB_TARGET_IS_SPARC64)
	{
	  elf32_greg_t tmp32;

	  read_register_gen (PS_REGNUM, (char *) &tmp32);
	  u.regs.tstate = psr_to_tstate_icc(tmp32);

	  read_register_gen (PC_REGNUM, (char *) &tmp32);
	  u.regs.tpc = tmp32;

	  read_register_gen (NPC_REGNUM, (char *) &tmp32);
	  u.regs.tnpc = tmp32;

	  read_register_gen (Y_REGNUM, (char *) &tmp32);
	  u.regs.y = tmp32;
	}
      else
	{
	  elf64_greg_t ccr;

	  read_register_gen(CCR_REGNUM, (char *) &ccr);
	  u.regs.tstate = ccr_to_tstate(ccr);

	  read_register_gen(PC_REGNUM, (char *) &u.regs.tpc);
	  read_register_gen(NPC_REGNUM, (char *) &u.regs.tnpc);
	  read_register_gen(Y_REGNUM, (char *) &u.regs.y);
	}

      if (0 != ptrace (PTRACE_SETREGS, proc_id,
		       (PTRACE_ARG3_TYPE) &u.regs, 0))
	perror ("ptrace_setregs");

      for (i = 0; i < 32; i++)
	read_register_gen (FP0_REGNUM + i, (char *) &u.fpregs.regs[i]);
      if (GDB_TARGET_IS_SPARC64)
	{
	  elf64_greg_t *fp_double = (elf64_greg_t *) &u.fpregs.regs[32];
	  for (i = 0; i < 16; i++)
	    read_register_gen (FP0_REGNUM + 32 + i, (char *) &fp_double[i]);
	  read_register_gen (FSR_REGNUM, (char *) &u.fpregs.fsr);
	}
      else
	read_register_gen (FPS_REGNUM, ((char *) &u.fpregs.fsr) + offset);

      if (0 != ptrace (PTRACE_SETFPREGS, proc_id,
		       (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	perror ("ptrace_setfpregs");
    }
  else
    {
      union {
	struct pt32_regs regs;
	struct pt32_fpregs fpregs;
      } u;

      for (i = G1_REGNUM; i < G0_REGNUM + 16; i++)
	{
	  char *regptr = ((char *)&u.regs.u_regs[i - G1_REGNUM + UREG_G1]);

	  read_register_gen (i, regptr);
	}
      read_register_gen(PS_REGNUM, (char *) &u.regs.psr);
      read_register_gen(PC_REGNUM, (char *) &u.regs.pc);
      read_register_gen(NPC_REGNUM, (char *) &u.regs.npc);
      read_register_gen(Y_REGNUM, (char *) &u.regs.y);

      if (0 != ptrace (PTRACE_SETREGS, proc_id,
		       (PTRACE_ARG3_TYPE) &u.regs, 0))
	perror ("ptrace_setregs");

      for (i = 0; i < 32; i++)
	read_register_gen (FP0_REGNUM + i, (char *) &u.fpregs.regs[i]);
      read_register_gen (FPS_REGNUM, (char *) &u.fpregs.fsr);

      if (0 != ptrace (PTRACE_SETFPREGS, proc_id,
		       (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	perror ("ptrace_setfpregs");
    }
}

void
store_inferior_registers (int regno)
{
  int intreg_size = SPARC_INTREG_SIZE;
  int proc_id, i, store_mask;
  char *buf;

  /* GNU/Linux LWP ID's are process ID's.  */
  if ((proc_id = TIDGET (inferior_ptid)) == 0)
    proc_id = PIDGET (inferior_ptid);		/* Not a threaded program.  */

  if (regno == G1_REGNUM)
    return;

  if (regno < 0 || regno == SP_REGNUM)
    {
      CORE_ADDR sp = read_sp ();

      if (!register_cached(L0_REGNUM + 5))
	internal_error (__FILE__, __LINE__, "failed internal consistency check");
      target_write_memory (sp,
			   &registers[REGISTER_BYTE (L0_REGNUM)],
			   16 * intreg_size);
    }
  else if (regno >= L0_REGNUM && regno <= I7_REGNUM)
    {
      CORE_ADDR sp = read_sp ();

      if (!register_cached(regno))
	internal_error (__FILE__, __LINE__, "failed internal consistency check");
      target_write_memory (sp + REGISTER_BYTE (regno) - REGISTER_BYTE (L0_REGNUM),
			   &registers[REGISTER_BYTE (regno)],
			   intreg_size);
      return;
    }

  if (regno < 0)
    {
      store_all_registers (proc_id);
      return;
    }

  if (sizeof (void *) == 8)
    {
      union {
	struct pt64_regs regs;
	struct pt64_fpregs fpregs;
      } u;
      int offset = (intreg_size == 4) ? 4 : 0;

      if (regno <= O7_REGNUM
	  || (regno >= PC_REGNUM && regno != FSR_REGNUM))
	{
	  if (0 != ptrace (PTRACE_GETREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.regs, 0))
	    perror ("ptrace_getregs");

	  if (regno >= G1_REGNUM && regno <= O7_REGNUM)
	    {
	      char *regptr = (char *) &u.regs.u_regs[regno - G1_REGNUM + UREG_G1];

	      if (offset != 0)
		memset (regptr, 0, sizeof(elf64_greg_t));
	      regptr += offset;
	      read_register_gen(regno, regptr);
	    }
	  if (! GDB_TARGET_IS_SPARC64)
	    {
	      elf32_greg_t tmp32;

	      if (regno == PS_REGNUM)
		{
		  read_register_gen (PS_REGNUM, (char *) &tmp32);
		  u.regs.tstate = psr_to_tstate_icc(tmp32);
		}
	      if (regno == PC_REGNUM)
		{
		  read_register_gen (PC_REGNUM, (char *) &tmp32);
		  u.regs.tpc = tmp32;
		}
	      if (regno == NPC_REGNUM)
		{
		  read_register_gen (NPC_REGNUM, (char *) &tmp32);
		  u.regs.tnpc = tmp32;
		}
	      if (regno == Y_REGNUM)
		{
		  read_register_gen (Y_REGNUM, (char *) &tmp32);
		  u.regs.y = tmp32;
		}
	    }
	  else
	    {
	      if (regno == CCR_REGNUM)
		{
		  elf64_greg_t ccr;

		  read_register_gen (CCR_REGNUM, (char *) &ccr);
		  u.regs.tstate = ccr_to_tstate(ccr);
		}
	      if (regno == PC_REGNUM)
		read_register_gen (PC_REGNUM, (char *) &u.regs.tpc);
	      if (regno == NPC_REGNUM)
		read_register_gen (NPC_REGNUM, (char *) &u.regs.tnpc);
	      if (regno == Y_REGNUM)
		read_register_gen (Y_REGNUM, (char *) &u.regs.y);
	    }

	  if (0 != ptrace (PTRACE_SETREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.regs, 0))
	    perror ("ptrace_setregs");
	}
      else
	{
	  if (0 != ptrace (PTRACE_GETFPREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	    perror ("ptrace_getfpregs");

	  if (GDB_TARGET_IS_SPARC64)
	    {
	      if (regno >= FP0_REGNUM && regno < FP0_REGNUM + 32)
		read_register_gen (regno, (char *) &u.fpregs.regs[regno - FP0_REGNUM]);
	      else if (regno >= FP0_REGNUM + 32 && regno < FP0_REGNUM + 48)
		{
		  elf64_greg_t *fp_double = (elf64_greg_t *) &u.fpregs.regs[32];
		  read_register_gen (regno,
				     (char *) &fp_double[regno - (FP0_REGNUM + 32)]);
		}
	      else if (regno == FSR_REGNUM)
		read_register_gen (FSR_REGNUM, (char *) &u.fpregs.fsr);
	    }
	  else
	    {
	      if (regno >= FP0_REGNUM && regno < FP0_REGNUM + 32)
		read_register_gen (regno, (char *) &u.fpregs.regs[regno - FP0_REGNUM]);
	      else if (regno == FPS_REGNUM)
		read_register_gen (FPS_REGNUM, ((char *) &u.fpregs.fsr) + offset);
	    }

	  if (0 != ptrace (PTRACE_SETFPREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	    perror ("ptrace_setfpregs");
	}
    }
  else
    {
      union {
	struct pt32_regs regs;
	struct pt32_fpregs fpregs;
      } u;

      if (regno <= O7_REGNUM
	  || (regno >= Y_REGNUM && regno != FPS_REGNUM))
	{
	  if (0 != ptrace (PTRACE_GETREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.regs, 0))
	    perror ("ptrace_getregs");

	  if (regno >= G1_REGNUM && regno <= O7_REGNUM)
	    {
	      char *regptr = (char *) &u.regs.u_regs[regno - G1_REGNUM + UREG_G1];

	      read_register_gen(regno, regptr);
	    }
	  if (regno == PS_REGNUM)
	    read_register_gen (PS_REGNUM, (char *) &u.regs.psr);
	  if (regno == PC_REGNUM)
	    read_register_gen (PC_REGNUM, (char *) &u.regs.pc);
	  if (regno == NPC_REGNUM)
	    read_register_gen (NPC_REGNUM, (char *) &u.regs.npc);
	  if (regno == Y_REGNUM)
	    read_register_gen (Y_REGNUM, (char *) &u.regs.y);

	  if (0 != ptrace (PTRACE_SETREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.regs, 0))
	    perror ("ptrace_setregs");
	}
      else
	{
	  if (0 != ptrace (PTRACE_GETFPREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	    perror ("ptrace_getfpregs");

	  if (regno >= FP0_REGNUM && regno < FP0_REGNUM + 32)
	    read_register_gen (regno, (char *) &u.fpregs.regs[regno - FP0_REGNUM]);
	  else if (regno == FPS_REGNUM)
	    read_register_gen (FPS_REGNUM, (char *) &u.fpregs.fsr);

	  if (0 != ptrace (PTRACE_SETFPREGS, proc_id,
			   (PTRACE_ARG3_TYPE) &u.fpregs, 0))
	    perror ("ptrace_setfpregs");
	}
    }
}

/* Register that we are able to handle ELF file formats using standard
   procfs "regset" structures.  */

static struct core_fns sparc_linux_core_fns =
{
  bfd_target_elf_flavour,		/* core_flavour */
  default_check_format,			/* check_format */
  default_core_sniffer,			/* core_sniffer */
  fetch_core_registers,			/* core_read_registers */
  NULL					/* next */
};

void
_initialize_core_sparc_linux (void)
{
  add_core_fns (&sparc_linux_core_fns);
}

/* Oh well, glibc's Linux ptrace() implementation does not understand
   the 5-arg calls that we need here.  */
#define __NR_ptrace_syscall __NR_ptrace
_syscall5(int,ptrace_syscall,int,req,int,pid,unsigned long,memaddr,int,len,char *,myaddr)

/* PTRACE_{READ,WRITE}DATA works under Sparc GNU/Linux and it is sooooo much
   faster than what infptrace.c does.  For example, using infptrace.c
   the test gdb.base/huge.exp can take forever, with this code it is nearly
   instantaneous.  */

int
child_xfer_memory (CORE_ADDR memaddr, char *myaddr, int len, int write,
		   struct mem_attrib *attrib, struct target_ops *target)
{
  int proc_id, err;

  if ((proc_id = TIDGET (inferior_ptid)) == 0)
    proc_id = PIDGET (inferior_ptid);		/* Not a threaded program.  */

  errno = 0;

  if (write)
    {
      ptrace_syscall (PTRACE_WRITEDATA, proc_id, memaddr, len, myaddr);
    }
  else
    {
      ptrace_syscall (PTRACE_READDATA, proc_id, memaddr, len, myaddr);
    }

  if (errno)
    return 0;

  return len;
}
